# -*- coding: utf-8 -*-
"""
Test hooks in harvest
"""

from nose.tools import (
    assert_in,
    assert_not_in,
)

from tests.util import in_directory, run_scenario


@in_directory(__file__, 'django', 'couves')
def test_django_agains_couves():
    'it always call @after.all hooks, even after exceptions'

    _, out = run_scenario(**{'-s': None})

    assert_in("Couves before all", out)
    assert_in("Couves after all", out)


@in_directory(__file__, 'django', 'couves')
def test_django_agains_couves_nohooks():
    """
    it only calls @before.all and @after.all hooks if there are features found
    """

    _, out = run_scenario(**{
        '--tags': 'nothingwillbefound',
        '-s': None,
    })

    assert_not_in("Couves before all", out)
    assert_not_in("Couves after all", out)



#

"""
Email backend that sends mails to a multiprocessing queue
"""
from aloe_django import mail
from django.core.mail.backends.base import BaseEmailBackend


class QueueEmailBackend(BaseEmailBackend):
    """
    Email backend that allows mail to be captured by tests run between
    separate processes via :mod:`multiprocessing`.
    """

    def send_messages(self, messages):
        for message in messages:
            mail.queue.put(message)  # pylint:disable=no-member

        return len(messages)



# -*- coding: utf-8 -*-
#

"""
Django integration for Aloe
"""

from django.core.exceptions import ImproperlyConfigured

try:
    import django.test

    from aloe.testclass import TestCase as AloeTestCase

    class TestCase(django.test.LiveServerTestCase, AloeTestCase):
        """
        Base test class for Django Gherkin tests.

        Inherits from both :class:`aloe.testclass.TestCase` and
        :class:`django.test.LiveServerTestCase`.
        """

        pass

except (ImproperlyConfigured, ImportError):
    # Probably running tests for Aloe-Django and Django isn't configured
    pass



# -*- coding: utf-8 -*-
"""
Test the harvest command.
"""

from nose.tools import assert_equals  # pylint:disable=no-name-in-module
from tests.util import getstatusoutput, in_directory


@in_directory(__file__, 'django', 'alfaces')
def test_django_agains_alfaces():
    'running the "harvest" django command with verbosity 3'

    status, out = getstatusoutput(
        "python manage.py harvest --verbosity=3")
    assert_equals(status, 0, out)

    assert "Test the django app DO NOTHING" in out
    assert "Test the django app FOO BAR" in out


@in_directory(__file__, 'django', 'alfaces')
def test_running_only_specified_features():
    'it can run only the specified features, passing the file path'

    status, out = getstatusoutput(
        "python manage.py harvest --verbosity=3 "
        "foobar/features/foobar.feature")

    assert_equals(status, 0, out)

    assert "Test the django app FOO BAR" in out
    assert "Test the django app DO NOTHING" not in out


@in_directory(__file__, 'django', 'alfaces')
def test_specifying_features_in_inner_directory():
    'it can run only the specified features from a subdirectory'

    status, out = getstatusoutput(
        "python manage.py harvest --verbosity=3 "
        "foobar/features/deeper/deeper/leaf.feature")

    assert_equals(status, 0, out)

    assert "Test the django app FOO BAR" not in out
    assert "Test a feature in an inner directory" in out
    assert "Test the django app DO NOTHING" not in out



#

"""
Step definitions and utilities for working with Django models.
"""

from __future__ import print_function
# pylint:disable=redefined-builtin
from builtins import str
# pylint:disable=redefined-builtin

import warnings

from django.core.management.color import no_style
from django.db import connection
from django.db.models.loading import get_models
from functools import partial

from aloe import step
from aloe.tools import guess_types

__all__ = ('writes_models', 'write_models',
           'tests_existence', 'test_existence',
           'reset_sequence')


STEP_PREFIX = r'(?:Given|And|Then|When) '


def _models_generator():
    """
    Build a hash of model verbose names to models
    """
    for model in get_models():
        yield (str(model._meta.verbose_name), model)
        yield (str(model._meta.verbose_name_plural), model)


try:
    MODELS = dict(_models_generator())
except:  # pylint:disable=bare-except
    warnings.warn("Models not loaded!")


_WRITE_MODEL = {}


def writes_models(model):
    """
    Register a model-specific create and update function.

    This can then be accessed via the steps:

    .. code-block:: gherkin

        And I have foos in the database:
            | name | bar  |
            | Baz  | Quux |

        And I update existing foos by pk in the database:
            | pk | name |
            | 1  | Bar  |

    A method for a specific model can define a function ``write_badgers(data,
    field)``, which creates and updates the Badger model and decorating it with
    the ``writes_models(model_class)`` decorator:

    .. code-block:: python

        @writes_models(Profile)
        def write_profile(data, field):
            '''Creates a Profile model'''

            for hash_ in data:
                if field:
                    profile = Profile.objects.get(**{field: hash_[field]})
                else:
                    profile = Profile()

                ...

            reset_sequence(Profile)

    The function must accept a list of data hashes and a field name. If field
    is not None, it is the field that must be used to get the existing objects
    out of the database to update them; otherwise, new objects must be created
    for each data hash.

    Follow up model creation with a call to :func:`reset_sequence` to
    update the database sequences.

    If you only want to modify the hash, you can make modifications and then
    pass it on to :func:`write_models`.

    .. code-block:: python

        @writes_models(Profile)
        def write_profile(data, field):
            '''Creates a Profile model'''

            for hash_ in data:

                # modify hash

            return write_models(Profile, data, field)
    """

    def decorated(func):
        """
        Decorator for the creation function.
        """
        _WRITE_MODEL[model] = func
        return func

    return decorated


_TEST_MODEL = {}


def tests_existence(model):
    """
    Register a model-specific existence test.

    This can then be accessed via the steps:

    .. code-block:: gherkin

        Then foos should be present in the database:
            | name   | bar |
            | badger | baz |

        Then foos should not be present in the database:
            | name   | bar |
            | badger | baz |

    A method for a specific model can define a function
    ``test_badgers(queryset, data)`` and decorating it with the
    ``tests_existence(model_class)`` decorator:

    .. code-block:: python

        @tests_existence(Profile)
        def test_profile(queryset, data):
            '''Test a Profile model'''

            # modify data ...

            return test_existence(queryset, data)

    If you only want to modify the hash, you can make modifications then pass
    it on to test_existence().
    """

    def decorated(func):
        """
        Decorator for the existence function.
        """
        _TEST_MODEL[model] = func
        return func

    return decorated


def get_model(model):
    """
    Convert a model's verbose name to the model class. This allows us to
    use the models verbose name in steps.
    """

    name = model.lower()
    model = MODELS.get(model, None)

    assert model, "Could not locate model by name '%s'" % name

    return model


def reset_sequence(model):
    """
    Reset the ID sequence for a model.
    """
    sql = connection.ops.sequence_reset_sql(no_style(), [model])
    for cmd in sql:
        connection.cursor().execute(cmd)


def _dump_model(model, attrs=None):
    """
    Dump the model fields for debugging.
    """

    fields = []

    for field in model._meta.fields:
        fields.append((field.name, str(getattr(model, field.name))))

    if attrs is not None:
        for attr in attrs:
            fields.append((attr, str(getattr(model, attr))))

    for field in model._meta.many_to_many:
        vals = getattr(model, field.name)
        fields.append(field.name, '{val} ({count})'.format(
            val=', '.join(map(str, vals.all())),
            count=vals.count(),
        ))

    print(', '.join(
        '{0}={1}'.format(field, value)
        for field, value in fields
    ))


def test_existence(queryset, data):
    """
    :param queryset: a Django queryset
    :param data: a single model to check for
    :returns: True if the model exists

    Test existence of a given hash in a `queryset` (or among all model
    instances if a model is given).

    Useful when registering custom tests with :func:`tests_existence`.
    """

    fields = {}
    extra_attrs = {}
    for key, value in data.items():
        if key.startswith('@'):
            # this is an attribute
            extra_attrs[key[1:]] = value
        else:
            fields[key] = value

    filtered = queryset.filter(**fields)

    if filtered.exists():
        return any(
            all(getattr(obj, k) == v for k, v in extra_attrs.items())
            for obj in filtered.all()
        )

    return False


def _model_exists_step(self, model, should_exist):
    """
    Test for the existence of a model matching the given data.
    """

    model = get_model(model)
    data = guess_types(self.hashes)

    queryset = model.objects

    try:
        existence_check = _TEST_MODEL[model]
    except KeyError:
        existence_check = test_existence

    failed = 0
    try:
        for hash_ in data:
            match = existence_check(queryset, hash_)

            if should_exist:
                assert match, \
                    "%s does not exist: %s" % (model.__name__, hash_)
            else:
                assert not match, \
                    "%s exists: %s" % (model.__name__, hash_)

    except AssertionError as exc:
        print(exc)
        failed += 1

    if failed:
        print("Rows in DB are:")
        for model in queryset.all():
            _dump_model(model,
                        attrs=[k[1:]
                               for k in data[0].keys()
                               if k.startswith('@')])

        if should_exist:
            raise AssertionError("%i rows missing" % failed)
        else:
            raise AssertionError("%i rows found" % failed)


@step(STEP_PREFIX +
      r'(?:an? )?([A-Z][a-z0-9_ ]*) should be present in the database')
def _model_exists_positive_step(self, model):
    """
    Test for the existence of a model matching the given data.

    Column names are included in a query to the database. To check model
    attributes that are not database columns (i.e. properties) prepend the
    column with an ``@`` sign.

    Example:

    .. code-block:: gherkin

        Then foos should be present in the database:
            | name   | @bar |
            | badger | baz  |

    See :func:`tests_existence`.
    """
    return _model_exists_step(self, model, True)


@step(STEP_PREFIX +
      r'(?:an? )?([A-Z][a-z0-9_ ]*) should not be present in the database')
def _model_exists_negative_step(self, model):
    """
    Tests for the existence of a model matching the given data.

    Column names are included in a query to the database. To check model
    attributes that are not database columns (i.e. properties). Prepend the
    column with an ``@`` sign.

    Example:

    .. code-block:: gherkin

        Then foos should not be present in the database:
            | name   | @bar |
            | badger | baz  |

    See :func:`tests_existence`.
    """
    return _model_exists_step(self, model, False)


def write_models(model, data, field):
    """
    :param model: a Django model class
    :param data: a list of hashes to build models from
    :param field: a field name to match models on, or None
    :returns: a list of models written

    Create or update models for each data hash.

    `field` is the field that is used to get the existing models out of
    the database to update them; otherwise, if ``field=None``, new models are
    created.

    Useful when registering custom tests with :func:`writes_models`.
    """
    written = []

    for hash_ in data:
        if field:
            if field not in hash_:
                raise KeyError(("The \"%s\" field is required for all update "
                                "operations") % field)

            model_kwargs = {field: hash_[field]}
            model_obj = model.objects.get(**model_kwargs)

            for to_set, val in hash_.items():
                setattr(model_obj, to_set, val)

            model_obj.save()

        else:
            model_obj = model.objects.create(**hash_)

        written.append(model_obj)

    reset_sequence(model)
    return written


def _write_models_step(self, model, field=None):
    """
    Write or update a model.
    """

    model = get_model(model)
    data = guess_types(self.hashes)

    try:
        func = _WRITE_MODEL[model]
    except KeyError:
        func = partial(write_models, model)

    func(data, field)


@step(r'I have(?: an?)? ([a-z][a-z0-9_ ]*) in the database:')
def _write_models_step_new(*args):
    """
    Create models in the database.

    Syntax:

        I have `model` in the database:

    Example:

    .. code-block:: gherkin

        And I have foos in the database:
            | name | bar  |
            | Baz  | Quux |

    See :func:`writes_models`.
    """
    return _write_models_step(*args)


@step(r'I update(?: an?)? existing ([a-z][a-z0-9_ ]*) by ([a-z][a-z0-9_]*) '
      'in the database:')
def _write_models_step_update(*args):
    """
    Update existing models in the database, specifying a column to match on.

    Syntax:

        I update `model` by `key` in the database:

    Example:

    .. code-block:: gherkin

        And I update existing foos by pk in the database:
            | pk | name |
            | 1  | Bar  |

    See :func:`writes_models`.
    """
    return _write_models_step(*args)


@step(STEP_PREFIX + r'([A-Z][a-z0-9_ ]*) with ([a-z]+) "([^"]*)"' +
      r' has(?: an?)? ([A-Z][a-z0-9_ ]*) in the database:')
def _create_models_for_relation_step(self, rel_model_name,
                                     rel_key, rel_value, model):
    """
    Create a new model linked to the given model.

    Syntax:

        And `model` with `field` "`value`" has `new model` in the database:

    Example:

    .. code-block:: gherkin

        And project with name "Ball Project" has goals in the database:
            | description                             |
            | To have fun playing with balls of twine |
    """

    model = get_model(model)
    lookup = {rel_key: rel_value}
    rel_model = get_model(rel_model_name).objects.get(**lookup)

    data = guess_types(self.hashes)

    for hash_ in data:
        hash_['%s' % rel_model_name] = rel_model

    try:
        func = _WRITE_MODEL[model]
    except KeyError:
        func = partial(write_models, model)

    func(data, None)


@step(STEP_PREFIX + r'([A-Z][a-z0-9_ ]*) with ([a-z]+) "([^"]*)"' +
      r' is linked to ([A-Z][a-z0-9_ ]*) in the database:')
def _create_m2m_links_step(self, rel_model_name,
                           rel_key, rel_value, relation_name):
    """
    Link many-to-many models together.

    Syntax:

        And `model` with `field` "`value`" is linked to `other model` in the
        database:

    Example:

    .. code-block:: gherkin

        And article with name "Guidelines" is linked to tags in the database:
            | name   |
            | coding |
            | style  |
    """

    lookup = {rel_key: rel_value}
    rel_model = get_model(rel_model_name).objects.get(**lookup)
    relation = None
    for m2m in rel_model._meta.many_to_many:
        if relation_name in (m2m.name, m2m.verbose_name):
            relation = getattr(rel_model, m2m.name)
            break
    if not relation:
        try:
            relation = getattr(rel_model, relation_name)
        except AttributeError:
            pass
    assert relation, \
        "%s does not have a many-to-many relation named '%s'" % (
            rel_model._meta.verbose_name.capitalize(),
            relation_name,
        )
    m2m_model = relation.model

    for hash_ in self.hashes:
        relation.add(m2m_model.objects.get(**hash_))


@step(r'There should be (\d+) ([a-z][a-z0-9_ ]*) in the database')
def _model_count_step(self, count, model):
    """
    Count the number of models in the database.

    Example:

    .. code-block:: gherkin

        Then there should be 0 goals in the database
    """

    model = get_model(model)

    expected = int(count)
    found = model.objects.count()
    assert found == expected, "Expected %d %s, found %d." % \
        (expected, model._meta.verbose_name_plural, found)



# -*- coding: utf-8 -*-
"""
Test mail steps.
"""

from tests.util import in_directory, run_scenario

from nose.tools import (  # pylint:disable=no-name-in-module
    assert_equals,
    assert_in,
    assert_not_equals,
)


@in_directory(__file__, 'django', 'bamboo')
def test_mail_count():
    'Mail count is checked through Lettuce steps'

    status, out = run_scenario('leaves', 'count', 1)
    assert_equals(status, 0, out)
    status, out = run_scenario('leaves', 'count', 2)
    assert_equals(status, 0, out)

    status, out = run_scenario('leaves', 'count', 3)
    assert_not_equals(status, 0)
    assert_in("Length of outbox is 1", out)


@in_directory(__file__, 'django', 'bamboo')
def test_mail_content():
    'Mail content is checked through Lettuce steps'

    status, out = run_scenario('leaves', 'content', 1)
    assert_equals(status, 0, out)
    status, out = run_scenario('leaves', 'content', 2)
    assert_equals(status, 0, out)
    status, out = run_scenario('leaves', 'content', 3)
    assert_equals(status, 0, out)

    status, out = run_scenario('leaves', 'content', 4)
    assert_not_equals(status, 0)
    assert_in("An email contained expected text in the body", out)

    status, out = run_scenario('leaves', 'content', 5)
    assert_not_equals(status, 0)
    assert_in("No email contained the HTML", out)


@in_directory(__file__, 'django', 'bamboo')
def test_mail_fail():
    'Mock mail failure dies with error'

    status, out = run_scenario('leaves', 'mock-failure', 1)
    assert_not_equals(status, 0)
    assert_in("SMTPException: Failure mocked by aloe", out)



"""
Step definitions for working with Django email.
"""
from __future__ import print_function

from smtplib import SMTPException

from django.core import mail
from django.test.html import parse_html

from aloe import step
from nose.tools import assert_in  # pylint:disable=no-name-in-module

__all__ = ()


STEP_PREFIX = r'(?:Given|And|Then|When) '
CHECK_PREFIX = r'(?:And|Then) '
EMAIL_PARTS = ('subject', 'body', 'from_email', 'to', 'bcc', 'cc')
GOOD_MAIL = mail.EmailMessage.send


@step(CHECK_PREFIX + r'I have sent (\d+) emails?')
def mail_sent_count(self, count):
    """
    Test that `count` mails have been sent.

    Syntax:

        I have sent `count` emails

    Example:

    .. code-block:: gherkin

        Then I have sent 2 emails
    """
    count = int(count)
    assert len(mail.outbox) == count, "Length of outbox is {0}".format(count)


@step(r'I have not sent any emails')
def mail_not_sent(self):
    """
    Test no emails have been sent.

    Example:

    .. code-block:: gherkin

        Then I have not sent any emails
    """
    return mail_sent_count(self, 0)


@step(CHECK_PREFIX + (r'I have sent an email with "([^"]*)" in the ({0})'
                      '').format('|'.join(EMAIL_PARTS)))
def mail_sent_content(self, text, part):
    """
    Test an email contains (assert text in) the given text in the relevant
    message part (accessible as an attribute on the email object).

    This step strictly applies whitespace.

    Syntax:

        I have sent an email with "`text`" in the `part`

    Example:

    .. code-block:: gherkin

        Then I have sent an email with "pandas" in the body
    """
    assert any(text in getattr(email, part)
               for email
               in mail.outbox), \
        "An email contained expected text in the {0}".format(part)


@step(CHECK_PREFIX + r'I have sent an email with the following in the body:')
def mail_sent_content_multiline(self):
    """
    Test the body of an email contains (assert text in) the given multiline
    string.

    This step strictly applies whitespace.

    Example:

    .. code-block:: gherkin

        Then I have sent an email with the following in the body:
        \"\"\"
        Dear Mr. Panda,
        \"\"\"
    """
    return mail_sent_content(self, self.multiline, 'body')


@step(CHECK_PREFIX +
      r'I have sent an email with the following HTML alternative:')
def mail_sent_contains_html(self):
    """
    Test that an email contains the HTML (assert HTML in) in the multiline as
    one of its MIME alternatives.

    The HTML is normalised by passing through Django's
    :func:`django.test.html.parse_html`.

    Example:

    .. code-block:: gherkin

        And I have sent an email with the following HTML alternative:
        \"\"\"
        <p><strong>Name:</strong> Sir Panda</p>
        <p><strong>Phone:</strong> 0400000000</p>
        <p><strong>Email:</strong> sir.panda@pand.as</p>
        \"\"\"
    """

    for email in mail.outbox:
        try:
            html = next(content for content, mime in email.alternatives
                        if mime == 'text/html')
            dom1 = parse_html(html)
            dom2 = parse_html(self.multiline)

            assert_in(dom1, dom2)

        except AssertionError as exc:
            print("Email did not match", exc)
            # we intentionally eat the exception
            continue

        return True

    raise AssertionError("No email contained the HTML")


@step(STEP_PREFIX + r'I clear my email outbox')
def mail_clear(self):
    """
    Clear the email outbox.

    Example:

    .. code-block:: gherkin

        Given I clear my email outbox
    """
    mail.EmailMessage.send = GOOD_MAIL
    mail.outbox = []


def broken_send(*args, **kwargs):
    """
    Broken send function for email_broken step
    """
    raise SMTPException("Failure mocked by aloe_django")


@step(STEP_PREFIX + r'sending email does not work')
def email_broken(self):
    """
    Cause sending email to raise an exception.

    This allows simulating email failure.

    Example:

    .. code-block:: gherkin

        Given sending email does not work
    """
    mail.EmailMessage.send = broken_send
